/**
 * Back 2 Browser Bytecode Translator
 * Copyright (C) 2012 Jaroslav Tulach <jaroslav.tulach@apidesign.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. Look for COPYING file in the top folder.
 * If not, see http://opensource.org/licenses/GPL-2.0.
 */
package org.apidesign.bck2brwsr.htmlpage;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PushbackInputStream;
import java.io.Reader;
import java.net.URL;
import java.util.Iterator;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;
import javafx.application.Platform;
import netscape.javascript.JSObject;
import org.apidesign.bck2brwsr.core.JavaScriptBody;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.json.JSONTokener;

/**
 *
 * @author Jaroslav Tulach <jtulach@netbeans.org>
 */
public final class ConvertTypes {
    private static final Logger LOG = Logger.getLogger(ConvertTypes.class.getName());
    ConvertTypes() {
    }
    
    public static String toString(Object object, String property) {
        Object ret = getProperty(object, property);
        return ret == null ? null : ret.toString();
    }

    public static double toDouble(Object object, String property) {
        Object ret = getProperty(object, property);
        return ret instanceof Number ? ((Number)ret).doubleValue() : Double.NaN;
    }

    public static int toInt(Object object, String property) {
        Object ret = getProperty(object, property);
        return ret instanceof Number ? ((Number)ret).intValue() : Integer.MIN_VALUE;
    }

    public static <T> T toModel(Class<T> modelClass, Object object, String property) {
        Object ret = getProperty(object, property);
        if (ret instanceof JSObject) {
            Object real = ((JSObject)ret).getMember("koModel");
            if (real != null) {
                ret = real;
            }
        }
        if (ret == null || modelClass.isInstance(ret)) {
            return modelClass.cast(ret);
        }
        throw new IllegalStateException("Value " + ret + " is not of type " + modelClass);
    }
    
    public static String toJSON(Object value) {
        if (value == null) {
            return "null";
        }
        if (value instanceof String) {
            return '"' + 
                ((String)value).
                    replace("\"", "\\\"").
                    replace("\n", "\\n").
                    replace("\r", "\\r").
                    replace("\t", "\\t")
                + '"';
        }
        return value.toString();
    }
    
    @JavaScriptBody(args = { "object", "property" },
        body = "if (property === null) return object;\n"
        + "var p = object[property]; return p ? p : null;"
    )
    private static Object getProperty(Object object, String property) {
        if (property == null) {
            return object;
        } else {
            if (object instanceof JSObject) {
                JSObject jo = (JSObject)object;
                return jo.getMember(property);
            }
            return null;
        }
    }
    
    public static String createJSONP(Object[] jsonResult, Runnable whenDone) {
        return "json" + Integer.toHexString(whenDone.hashCode());
    }

    public static void loadJSON(
        String url, Object[] jsonResult, Runnable whenDone, String jsonp
    ) {
        REQ.execute(new LoadJSON(url, jsonResult, whenDone, jsonp));
    }
    
    private static final Executor REQ = Executors.newCachedThreadPool();
    private static final class LoadJSON implements Runnable {
        
        private final String url;
        private final Object[] jsonResult;
        private final Runnable whenDone;
        private final String jsonp;

        LoadJSON(String url, Object[] jsonResult, Runnable whenDone, String jsonp) {
            this.url = url;
            this.jsonResult = jsonResult;
            this.whenDone = whenDone;
            this.jsonp = jsonp;
        }

        @Override
        public void run() {
            if (Platform.isFxApplicationThread()) {
                whenDone.run();
                return;
            }
            try {
                URL u = new URL(url.replace(" ", "%20"));
                InputStream is = u.openStream();
                if (jsonp != null) {
                    PushbackInputStream pis = new PushbackInputStream(is, 1);
                    is = pis;
                    for (;;) {
                        int ch = pis.read();
                        if (ch == -1) {
                            break;
                        }
                        if (ch == '{') {
                            pis.unread(ch);
                            break;
                        }
                    }
                }
                Reader r = new InputStreamReader(is, "UTF-8");
                
                JSONTokener tok = new JSONTokener(r);
                JSONObject obj = new JSONObject(tok);
                jsonResult[0] = convertToArray(obj);
            } catch (JSONException | IOException ex) {
                jsonResult[0] = ex;
            } finally {
                Platform.runLater(this);
            }
        }
        
        private static Object convertToArray(Object o) throws JSONException {
            if (o instanceof JSONArray) {
                JSONArray ja = (JSONArray)o;
                Object[] arr = new Object[ja.length()];
                for (int i = 0; i < arr.length; i++) {
                    arr[i] = convertToArray(ja.get(i));
                }
                return arr;
            } else if (o instanceof JSONObject) {
                JSONObject obj = (JSONObject)o;
                Iterator it = obj.keys();
                while (it.hasNext()) {
                    String key = (String)it.next();
                    obj.put(key, convertToArray(obj.get(key)));
                }
                return obj;
            } else {
                return o;
            }
        }
    }
    
    @JavaScriptBody(args = { "url", "jsonp" }, body = 
        "var scrpt = window.document.createElement('script');\n "
        + "scrpt.setAttribute('src', url);\n "
        + "scrpt.setAttribute('id', jsonp);\n "
        + "scrpt.setAttribute('type', 'text/javascript');\n "
        + "var body = document.getElementsByTagName('body')[0];\n "
        + "body.appendChild(scrpt);\n"
    )
    private static void loadJSONP(String url, String jsonp) {
        
    }
    
    public static void extractJSON(Object jsonObject, String[] props, Object[] values) {
        if (jsonObject instanceof JSONObject) {
            JSONObject obj = (JSONObject)jsonObject;
            for (int i = 0; i < props.length; i++) {
                try {
                    values[i] = obj.get(props[i]);
                } catch (JSONException ex) {
                    LOG.log(Level.SEVERE, "Can't read " + props[i] + " from " + jsonObject, ex);
                }
            }
            
        }
        
    }
    
}
